/*
 * Copyright (C) 2014 Red Hat, Inc.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the licence, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, see <http://www.gnu.org/licenses/>.
 *
 * Author: Matthew Barnes <mbarnes@redhat.com>
 */

#include "config.h"

#include "fcmdr-gsettings-backend.h"

#include <errno.h>
#include <pwd.h>
#include <string.h>
#include <unistd.h>

#include "fcmdr-extensions.h"
#include "fcmdr-generated.h"
#include "fcmdr-service.h"
#include "fcmdr-utils.h"

#define FCMDR_GSETTINGS_BACKEND_GET_PRIVATE(obj) \
	(G_TYPE_INSTANCE_GET_PRIVATE \
	((obj), FCMDR_TYPE_GSETTINGS_BACKEND, FCmdrGSettingsBackendPrivate))

#define FCMDR_GSETTINGS_PREAMBLE \
	"# Generated by " G_LOG_DOMAIN ". DO NOT EDIT.\n\n"

#define FCMDR_DCONF_DB_DIR "/etc/dconf/db"

struct _FCmdrGSettingsBackendPrivate {
	FCmdrDConf *dconf_interface;
};

/* Forward Declarations */
static void	fcmdr_gsettings_backend_initable_interface_init
						(GInitableIface *interface);

G_DEFINE_TYPE_WITH_CODE (
	FCmdrGSettingsBackend,
	fcmdr_gsettings_backend,
	FCMDR_TYPE_SERVICE_BACKEND,
	G_IMPLEMENT_INTERFACE (
		G_TYPE_INITABLE,
		fcmdr_gsettings_backend_initable_interface_init)
	fcmdr_ensure_extension_points_registered();
	g_io_extension_point_implement (
		FCMDR_SERVICE_BACKEND_EXTENSION_POINT_NAME,
		g_define_type_id,
		"org.gnome.gsettings", 0))

static gboolean
fcmdr_gsettings_backend_handle_write_profile_cb (FCmdrDConf *interface,
                                                 GDBusMethodInvocation *invocation,
                                                 const gchar *filename,
                                                 FCmdrGSettingsBackend *backend)
{
	GDBusConnection *connection;
	FCmdrService *service;
	const gchar *sender;
	GString *contents;
	GList *list, *link;
	uid_t uid = 0;
	struct passwd *pw;
	GError *local_error = NULL;

	service = fcmdr_service_backend_ref_service (
		FCMDR_SERVICE_BACKEND (backend));

	/* Find out the user ID of the sender. */

	connection = fcmdr_service_get_connection (service);
	sender = g_dbus_method_invocation_get_sender (invocation);

	fcmdr_get_connection_unix_user_sync (
		connection, sender, &uid, NULL, &local_error);

	if (local_error != NULL)
		goto exit;

	contents = g_string_new ("user-db:user\n");

	list = fcmdr_service_list_profiles_for_user (service, uid);

	for (link = list; link != NULL; link = g_list_next (link)) {
		FCmdrProfile *profile;
		const gchar *uid;

		profile = FCMDR_PROFILE (link->data);
		uid = fcmdr_profile_get_uid (profile);

		g_string_append_printf (
			contents, "system-db:fleet-commander-%s\n", uid);
	}

	g_file_set_contents (
		filename, contents->str, contents->len, &local_error);

	/* Transfer ownership of the written file to the caller.
	 * Failure here is not fatal to the WriteProfile method. */
	if ((pw = getpwuid (uid)) != NULL)
		chown (filename, pw->pw_uid, pw->pw_gid);

	g_list_free_full (list, (GDestroyNotify) g_object_unref);

	g_string_free (contents, TRUE);

exit:
	if (local_error == NULL) {
		fcmdr_dconf_complete_write_profile (interface, invocation);
	} else {
		g_dbus_method_invocation_take_error (invocation, local_error);
	}

	g_clear_object (&service);

	return TRUE;
}

static gboolean
fcmdr_gsettings_backend_write_key_file (GKeyFile *key_file,
                                        const gchar *filename,
                                        GError **error)
{
	GString *contents;
	gchar *data;
	gsize length = 0;
	gboolean success;

	contents = g_string_new (FCMDR_GSETTINGS_PREAMBLE);

	data = g_key_file_to_data (key_file, &length, NULL);
	g_string_append_len (contents, data, length);
	g_free (data);

	success = g_file_set_contents (
		filename, contents->str, contents->len, error);

	g_string_free (contents, TRUE);

	return success;
}

static gboolean
fcmdr_gsettings_backend_write_locks (GHashTable *locks,
                                     const gchar *filename,
                                     GError **error)
{
	GString *contents;
	GList *list, *link;
	gboolean success;

	contents = g_string_new (FCMDR_GSETTINGS_PREAMBLE);

	list = g_hash_table_get_keys (locks);
	list = g_list_sort (list, (GCompareFunc) strcmp);

	for (link = list; link != NULL; link = g_list_next (link)) {
		g_string_append (contents, link->data);
		g_string_append_c (contents, '\n');
	}

	g_list_free (list);

	success = g_file_set_contents (
		filename, contents->str, contents->len, error);

	g_string_free (contents, TRUE);

	return success;
}

static void
fcmdr_gsettings_backend_foreach_element_cb (JsonArray *json_array,
                                            guint index,
                                            JsonNode *json_node,
                                            gpointer user_data)
{
	GKeyFile *key_file = user_data;
	JsonObject *json_object;
	GVariant *value = NULL;
	const gchar *path;
	gchar *group;
	gchar *key;

	g_return_if_fail (JSON_NODE_HOLDS_OBJECT (json_node));

	json_object = json_node_get_object (json_node);
	g_return_if_fail (json_object_has_member (json_object, "key"));
	g_return_if_fail (json_object_has_member (json_object, "value"));

	path = json_object_get_string_member (json_object, "key");
	g_return_if_fail (path != NULL);

	/* Skip leading slashes. */
	while (*path == '/')
		path++;

	group = g_strdup (path);
	key = strrchr (group, '/');

	if (key != NULL)
		*key++ = '\0';

	json_node = json_object_get_member (json_object, "value");
	value = fcmdr_json_value_to_variant (json_node);

	/* This is how GKeyfileSettingsBackend does it. */
	if (key != NULL && value != NULL) {
		gchar *str = g_variant_print (value, FALSE);
		g_key_file_set_value (key_file, group, key, str);
		g_variant_unref (g_variant_ref_sink (value));
		g_free (str);
	}

	g_free (group);
}

static void
fcmdr_gsettings_backend_build_profiles_table (FCmdrProfile *profile,
                                              GHashTable *profiles_table)
{
	const gchar *uid;
	gchar *basename;
	gchar *filename;

	uid = fcmdr_profile_get_uid (profile);
	basename = g_strdup_printf ("fleet-commander-%s.d", uid);
	filename = g_build_filename (FCMDR_DCONF_DB_DIR, basename, NULL);

	/* Takes ownership of the filename string. */
	g_hash_table_replace (
		profiles_table, filename,
		g_object_ref (profile));

	g_free (basename);
}

static void
fcmdr_gsettings_backend_purge_old_profiles (GHashTable *profiles_table,
                                            const gchar *dirname)
{
	GDir *dir;
	const gchar *sysdb_name;
	GError *local_error = NULL;

	dir = g_dir_open (dirname, 0, &local_error);

	if (local_error != NULL) {
		g_critical (
			"Failed to open directory: %s: %s",
			dirname, local_error->message);
		g_error_free (local_error);
		return;
	}

	while ((sysdb_name = g_dir_read_name (dir)) != NULL) {
		gchar *sysdb_path;

		/* Make sure we only delete our own databases!
		 *
		 * This will delete both "fleet-commander-$UID"
		 * files and "fleet-commander-$UID.d" directories. */

		if (!g_str_has_prefix (sysdb_name, "fleet-commander-"))
			continue;

		sysdb_path = g_build_filename (dirname, sysdb_name, NULL);

		if (!g_hash_table_contains (profiles_table, sysdb_path)) {
			GFile *file;

			/* XXX Should use GFile API everywhere. */
			file = g_file_new_for_path (sysdb_path);
			fcmdr_recursive_delete_sync (file, NULL, &local_error);
			g_object_unref (file);

			if (local_error != NULL) {
				g_warning (
					"Failed to delete: %s: %s",
					sysdb_path, local_error->message);
				g_clear_error (&local_error);
			}
		}

		g_free (sysdb_path);
	}

	g_dir_close (dir);
}

static void
fcmdr_gsettings_backend_apply_one_profile (const gchar *sysdb_path,
                                           FCmdrProfile *profile,
                                           FCmdrServiceBackend *backend)
{
	JsonNode *json_node;
	JsonArray *json_array;
	GKeyFile *key_file;
	GHashTable *locks;
	gchar **groups;
	gsize ii, n_groups;
	gchar *locks_path;
	gchar *locks_file;
	gchar *sysdb_file;

	/* The profile may not have settings for this backend. */
	json_node = fcmdr_service_backend_dup_settings (backend, profile);
	if (json_node == NULL)
		return;

	g_return_if_fail (JSON_NODE_HOLDS_ARRAY (json_node));

	key_file = g_key_file_new ();

	locks = g_hash_table_new_full (
		(GHashFunc) g_str_hash,
		(GEqualFunc) g_str_equal,
		(GDestroyNotify) g_free,
		(GDestroyNotify) NULL);

	json_array = json_node_get_array (json_node);

	json_array_foreach_element (
		json_array,
		fcmdr_gsettings_backend_foreach_element_cb,
		key_file);

	groups = g_key_file_get_groups (key_file, &n_groups);

	for (ii = 0; ii < n_groups; ii++) {
		gchar **keys;
		gsize jj, n_keys;

		keys = g_key_file_get_keys (
			key_file, groups[ii], &n_keys, NULL);

		for (jj = 0; jj < n_keys; jj++) {
			g_hash_table_add (
				locks,
				g_strdup_printf (
				"/%s/%s", groups[ii], keys[jj]));
		}

		g_strfreev (keys);
	}

	g_strfreev (groups);

	sysdb_file = g_build_filename (sysdb_path, "generated", NULL);
	locks_path = g_build_filename (sysdb_path, "locks", NULL);
	locks_file = g_build_filename (locks_path, "generated", NULL);

	if (g_mkdir_with_parents (locks_path, 0755) == 0) {
		GError *local_error = NULL;

		fcmdr_gsettings_backend_write_key_file (
			key_file, sysdb_file, &local_error);

		if (local_error != NULL) {
			g_critical (
				"Failed to write file: %s: %s",
				sysdb_file, local_error->message);
			g_clear_error (&local_error);
		}

		fcmdr_gsettings_backend_write_locks (
			locks, locks_file, &local_error);

		if (local_error != NULL) {
			g_critical (
				"Failed to write file: %s: %s",
				locks_file, local_error->message);
			g_clear_error (&local_error);
		}

	} else {
		g_critical (
			"Failed to make directory: %s: %s",
			locks_path, g_strerror (errno));
	}

	json_node_free (json_node);

	g_key_file_free (key_file);
	g_hash_table_destroy (locks);

	g_free (sysdb_file);
	g_free (locks_path);
	g_free (locks_file);
}

static void
fcmdr_gsettings_backend_dispose (GObject *object)
{
	FCmdrGSettingsBackendPrivate *priv;

	priv = FCMDR_GSETTINGS_BACKEND_GET_PRIVATE (object);

	g_clear_object (&priv->dconf_interface);

	/* Chain up to parent's dispose() method. */
	G_OBJECT_CLASS (fcmdr_gsettings_backend_parent_class)->
		dispose (object);
}

static void
fcmdr_gsettings_backend_apply_profiles (FCmdrServiceBackend *backend,
                                        GList *profiles)
{
	if (g_mkdir_with_parents (FCMDR_DCONF_DB_DIR, 0755) == 0) {
		GHashTable *profiles_table;
		GError *local_error = NULL;

		profiles_table = g_hash_table_new_full (
			(GHashFunc) g_str_hash,
			(GEqualFunc) g_str_equal,
			(GDestroyNotify) g_free,
			(GDestroyNotify) g_object_unref);

		g_list_foreach (
			profiles,
			(GFunc) fcmdr_gsettings_backend_build_profiles_table,
			profiles_table);

		fcmdr_gsettings_backend_purge_old_profiles (
			profiles_table, FCMDR_DCONF_DB_DIR);

		g_hash_table_foreach (
			profiles_table,
			(GHFunc) fcmdr_gsettings_backend_apply_one_profile,
			backend);

		g_hash_table_destroy (profiles_table);

		g_spawn_command_line_sync (
			DCONF_PATH " update", NULL, NULL, NULL, &local_error);

		if (local_error != NULL) {
			g_critical (
				"Failed to spawn dconf tool: %s",
				local_error->message);
			g_clear_error (&local_error);
		}
	} else {
		g_critical (
			"Failed to make directory: %s: %s",
			FCMDR_DCONF_DB_DIR, g_strerror (errno));
	}
}

static gboolean
fcmdr_gsettings_backend_initable_init (GInitable *initable,
                                       GCancellable *cancellable,
                                       GError **error)
{
	FCmdrGSettingsBackendPrivate *priv;
	FCmdrService *service;
	GDBusConnection *connection;
	gboolean success;

	priv = FCMDR_GSETTINGS_BACKEND_GET_PRIVATE (initable);

	service = fcmdr_service_backend_ref_service (
		FCMDR_SERVICE_BACKEND (initable));

	connection = fcmdr_service_get_connection (service);

	success = g_dbus_interface_skeleton_export (
		G_DBUS_INTERFACE_SKELETON (priv->dconf_interface),
		connection, FCMDR_SERVICE_DBUS_OBJECT_PATH, error);

	g_object_unref (service);

	return success;
}

static void
fcmdr_gsettings_backend_class_init (FCmdrGSettingsBackendClass *class)
{
	GObjectClass *object_class;
	FCmdrServiceBackendClass *backend_class;

	g_type_class_add_private (
		class, sizeof (FCmdrGSettingsBackendPrivate));

	object_class = G_OBJECT_CLASS (class);
	object_class->dispose = fcmdr_gsettings_backend_dispose;

	backend_class = FCMDR_SERVICE_BACKEND_CLASS (class);
	backend_class->apply_profiles = fcmdr_gsettings_backend_apply_profiles;
}

static void
fcmdr_gsettings_backend_initable_interface_init (GInitableIface *interface)
{
	interface->init = fcmdr_gsettings_backend_initable_init;
}

static void
fcmdr_gsettings_backend_init (FCmdrGSettingsBackend *backend)
{
	backend->priv = FCMDR_GSETTINGS_BACKEND_GET_PRIVATE (backend);
	backend->priv->dconf_interface = fcmdr_dconf_skeleton_new ();

	g_signal_connect (
		backend->priv->dconf_interface,
		"handle-write-profile",
		G_CALLBACK (fcmdr_gsettings_backend_handle_write_profile_cb),
		backend);
}

